"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
 
"""

import numpy as np
import scipy.stats
from sklearn.metrics import mean_absolute_error


def perf_metrix(predictions: list, true_values: list):
    '''Compute various prediction performances metrics.'''
    
    pearson, pearson_p = scipy.stats.pearsonr(predictions, true_values)
    spearman, spearman_p = scipy.stats.spearmanr(predictions, true_values)
    mae = mean_absolute_error(predictions, true_values)
    
    pearson = float("{0:.3g}".format(pearson))
    pearson_p = float("{0:.3g}".format(pearson_p))
    spearman = float("{0:.3g}".format(spearman))
    spearman_p = float("{0:.3g}".format(spearman_p))
    mae = float("{0:.3g}".format(mae))
    std_ratio = float("{0:.3g}".format(np.std(predictions)/np.std(true_values)))
  
    return pearson, pearson_p, spearman, spearman_p, mae, std_ratio


def is_protein(seq, aa_list):
    """
    Check if a str corresponds to a protein sequence
    return bool
    """
    for aa in seq:
        if aa not in aa_list:
            return False
    return True



from setuptools import setup, find_packages
import os

here = os.path.abspath(os.path.dirname(__file__))

# Read dependencies from environment.yml
with open(os.path.join(here, 'environment.yml'), encoding='utf-8') as f:
    # We need to parse without yaml parser, since this runs before any packages are installed
    install_requires = []
    started = False
    for line in f:
        if started:
            install_requires.append(line.strip().replace('- ', ''))
        if 'pip:' in line:
            started = True
    if not install_requires:
        raise ValueError(f'Error parsing pip dependencies from environment.yml')

setup(
    name='nanomelt',
    version='1.1.0',
    description='A nanobody thermostability prediction tool',
    licence='CC BY-NC-SA 4.0',
    author='Aubin Ramon',
    author_email='ar2033@cam.ac.uk',
    long_description='See READ.me in GitLab repository https://gitlab.developers.cam.ac.uk/ch/sormanni/nanomelt.git',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'nanomelt=nanomelt.__main__:main'
        ]
    },
    include_package_data=True,
    package_data={'': ["model/saved_models/NanoMelt_finalmodel/*"]},
    install_requires=install_requires
)



"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0

 Predicting nanobody apparent melting temperatures with NanoMelt
"""

from .model.nanomelt import NanoMeltPredPipe
from .model.embedding.fasta_embed import alphabet

import argparse
import os,sys

from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq

from .model.modules.utils import is_protein





def run(args: argparse.Namespace):

    if args.do_align:
        import multiprocessing
        multiprocessing.set_start_method("fork", force=True)
        
    # Load the fasta file
    if os.path.isfile(args.input_filepath_or_seq):
        seq_records =  list(SeqIO.parse(args.input_filepath_or_seq, 'fasta'))
        if args.maximum_sequences_to_process > 0 and len(seq_records) > args.maximum_sequences_to_process :
            sys.stderr.write('\n**ERROR** Given args limit at maximum_sequences_to_process=%d, but input fasta file contains %d sequences!\n ABORTING RUN\n\n' % (args.maximum_sequences_to_process,len(seq_records)))
            return 1
    # If just sequence
    elif is_protein(args.input_filepath_or_seq, alphabet):
        seq_records = [SeqRecord(Seq(args.input_filepath_or_seq), id='single_seq')]

    # Run the model
    data = NanoMeltPredPipe(seq_records, args.do_align, args.ncpu, verbose=args.verbose)

    # Save data as .csv
    if not is_protein(args.input_filepath_or_seq, alphabet):
        data.to_csv(args.output_savefp, index=False)

    if is_protein(args.input_filepath_or_seq, alphabet):
        tm = list(data['NanoMelt Tm (C)'])[0]
        al_seq = list(data['Aligned Sequence'])[0]
        print(f'\n-> Sequence aligned as:\n')
        print(al_seq)
        print(f'\n\n---> Predicted apparent melting temperature -> {tm} (C)\n\n\n\n')
    else:
        print('\n\n---> NanoMelt prediction complete and results saved in %s\n\n\n\n' % (str(args.output_savefp).split('/')[-1]))


"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
 
"""


import pandas as pd
import csv
import os

import scipy.stats
import numpy as np
from sklearn.metrics import mean_squared_error

from .modules.strat_nest import StratifiedNestedCV
from .modules.ensemble import TopPred

if __name__ == "__main__":

    # Stratnested parameters
    strat_method = "expmeth_clust_kmedoids_blosum" # 'expmeth_clust_kmedoids_blosum' 'exp_method'
    regression = "ridge"
    num_outer_fold = 3
    num_inner_fold = 3
    random_states = [0,
                    42,
                    100
                     ]
    
    # Extract top predictions
    metric = "spearman" # or "pearson" or "mae"
    
    # Choose diverse embedding or regression
    do_diverse_sorting = True
    diverse_emb = True
    diverse_reg = False

    # Set number of features representing tm and exp method in the ensembled embedding
    num_total_temperatures = 8 # number of top temperatures selected for ensemble, need >= 1
    if do_diverse_sorting and diverse_reg: num_total_temperatures = 7

    # File paths
    saved_embed_path = "../../data/database_640.pkl"
    perf_csv_dir = "./saved_results/performances_exp_blosum_640_mae_no_mean"
    saved_predictions_path = "./saved_results/predictions_exp_blosum_640_mae_no_mean"   

    if not do_diverse_sorting: id = 'true_top'
    elif diverse_emb: id = 'div_emb'
    else: id = 'div_reg'

    if not os.path.exists(perf_csv_dir):
        os.makedirs(perf_csv_dir)
    fp_ensemb_metrics = os.path.join(perf_csv_dir, f"ens_no_mean_{metric}_{id}.csv")

    if not os.path.exists(saved_predictions_path):
        os.makedirs(saved_predictions_path)
    fp_predictions = os.path.join(saved_predictions_path, f"ens_no_mean_{metric}_{id}.csv")

    embed_df = pd.read_pickle(saved_embed_path)

    if os.path.exists(fp_ensemb_metrics): os.remove(fp_ensemb_metrics)
    if os.path.exists(fp_predictions): os.remove(fp_predictions)

    for nb_temp in range(1,num_total_temperatures+1):

        trial = f"tm_{nb_temp}"

        # Loop over all embeddings
        list_fin_pearson_train = []
        list_fin_spearman_train = []
        list_fin_RSME_train = []
        list_fin_SDR_train = []

        list_fin_pearson_test = []
        list_fin_spearman_test = []
        list_fin_RSME_test = []
        list_fin_SDR_test = []

        list_pred = []

        for seed_idx, seed in enumerate(random_states): 

            # Extract set of predicitons for ensembling
            data = TopPred(top_ranks=nb_temp, 
                                    strat_method=strat_method,
                                    metric=metric,
                                    fp_database=saved_embed_path,
                                    saved_predictions_path=saved_predictions_path,
                                    perf_csv_dir=perf_csv_dir,
                                    diverse_emb=diverse_emb,
                                    diverse_reg=diverse_reg,
                                    do_diverse_sorting=do_diverse_sorting,
                                    seed_idx=seed_idx)
            
            target = embed_df['tm'].values
            label = embed_df[f'{strat_method}'].values
            print("Full dataset: ", data.shape, target.shape, label.shape)

            # Save evaluation metrics
            train_metrics, test_metrics, list_predictions = StratifiedNestedCV(
                            data, target, label,
                            num_outer_fold, num_inner_fold, [seed], regression,
                            strat_method = strat_method,
                            return_averaged=False
                            )

            # Save perfs by seed
            list_fin_pearson_train.append(train_metrics[0])
            list_fin_spearman_train.append(train_metrics[3])
            list_fin_RSME_train.append(train_metrics[6])
            list_fin_SDR_train.append(train_metrics[8])

            list_fin_pearson_test.append(test_metrics[0])
            list_fin_spearman_test.append(test_metrics[3])
            list_fin_RSME_test.append(test_metrics[6])
            list_fin_SDR_test.append(test_metrics[8])

            list_pred.append(list_predictions[0])

            # GET BASELINE
            datatop = TopPred(top_ranks=1, 
                                strat_method=strat_method,
                                metric=metric,
                                fp_database=saved_embed_path,
                                saved_predictions_path=saved_predictions_path,
                                perf_csv_dir=perf_csv_dir,
                                diverse_emb=diverse_emb,
                                diverse_reg=diverse_reg,
                                do_diverse_sorting=do_diverse_sorting,
                                seed_idx=seed_idx)

            # Get performance top model
            pearson, pearson_p = scipy.stats.pearsonr(datatop[:,0], target)
            spearman, spearman_p = scipy.stats.spearmanr(datatop[:,0], target)
            mae = np.sqrt(mean_squared_error(datatop[:,0], target))
            print('[BASELINE] TOP PERFORMING PERF ON WHOLE')
            print('pearson: ', pearson, ' spearman: ', spearman, ' mae: ',mae)

        pearson_mean_train, pearson_std_train = np.mean(list_fin_pearson_train), np.std(list_fin_pearson_train)
        spearman_mean_train, spearman_std_train = np.mean(list_fin_spearman_train), np.std(list_fin_spearman_train)
        RSME_mean_train, RSME_std_train = np.mean(list_fin_RSME_train), np.std(list_fin_RSME_train)
        SDR_mean_train, SDR_std_train = np.mean(list_fin_SDR_train), np.std(list_fin_SDR_train)

        train_metrics = [pearson_mean_train, pearson_std_train, spearman_mean_train, spearman_std_train,
                         RSME_mean_train, RSME_std_train, SDR_mean_train, SDR_std_train]

        pearson_mean_test, pearson_std_test = np.mean(list_fin_pearson_test), np.std(list_fin_pearson_test)
        spearman_mean_test, spearman_std_test = np.mean(list_fin_spearman_test), np.std(list_fin_spearman_test)
        RSME_mean_test, RSME_std_test = np.mean(list_fin_RSME_test), np.std(list_fin_RSME_test)
        SDR_mean_test, SDR_std_test = np.mean(list_fin_SDR_test), np.std(list_fin_SDR_test)

        test_metrics = [pearson_mean_test, pearson_std_test, spearman_mean_test, spearman_std_test,
                         RSME_mean_test, RSME_std_test, SDR_mean_test, SDR_std_test]
  
        # Save metrics and predictions on whole dataset
        header = ['method','trial',
                'fin_pearson_train', 'std_pearson_train', 
                'fin_spearman_train', 'std_spearman_train',
                'fin_mae_train', 'std_mae_train', 
                'fin_std_ratio_train', 'std_std_ratio_train',
                'fin_pearson_test','std_pearson_test',
                'fin_spearman_test', 'std_spearman_test',
                'fin_mae_test','std_mae_test',  
                'fin_std_ratio_test', 'std_std_ratio_test',]

        metrics = []
        metrics.extend([regression, trial])
        metrics.extend(train_metrics)
        metrics.extend(test_metrics)
        
        if not os.path.exists(fp_ensemb_metrics):    
            with open(fp_ensemb_metrics, 'w', newline='') as f:
                csv_writer = csv.writer(f)
                csv_writer.writerow(header) 
                csv_writer.writerow(metrics)
        else:
            with open(fp_ensemb_metrics, 'a', newline='') as f:
                csv_writer = csv.writer(f)
                csv_writer.writerow(metrics)
                
        if not os.path.exists(fp_predictions):
            predictions = pd.DataFrame()
            predictions[f'{regression}_{trial}'] = ['_'.join([str(tm) for tm in preds]) for preds in np.array(list_pred).T]
            predictions.to_csv(fp_predictions,index=False,header=True)
        else:
            predictions = pd.read_csv(fp_predictions)
            predictions[f'{regression}_{trial}'] = ['_'.join([str(tm) for tm in preds]) for preds in np.array(list_pred).T]
            predictions.to_csv(fp_predictions,index=False,header=True)

            





"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0

 NanoMelt's model
"""


from .embedding.fasta_embed import embed_fasta

import numpy as np
import pandas as pd 
import os
import joblib

from tqdm import tqdm


import importlib.resources as pkg_resources
from sklearn.exceptions import InconsistentVersionWarning
import warnings

warnings.filterwarnings("ignore", category=InconsistentVersionWarning)

def predict(df_embed: pd.DataFrame, dict_submodels: dict,
			name_ens_model: str, folder_saved_models: str) -> list:
	'''
	Run the prediction from the ensemble model based on pre-computed embeded sequences.
	
	Parameters
    ----------
        - df_embed: pd.DataFrame
            Filepath to the fasta file of input sequences to compute the embeddings on.
        - dict_submodels: dict
            Dict of the pre-trained submodels. 
			Keys = name of saved submodel in folder_saved_models.
			Values = name of the embedding needed as inputs.
		- name_ens_model: str
			Name of the pre-trained ensemble model in folder_saved_models.
		- folder_saved_models: str
            Filepath tot he fodler with the saved models.

    Returns
    ----------
        - predictions: list
	'''
	
	# Predict for each embedding (retrained model)
	list_tms = []
	for sub_model in dict_submodels:
		fp_trained_model = os.path.join(folder_saved_models, sub_model + '.joblib')
		embeddings = df_embed[dict_submodels[sub_model]].values

		# Load model
		with open(fp_trained_model,"rb") as f:
			model = joblib.load(f)

		# Predict
		tms = model.predict(np.vstack(embeddings))
		assert len(tms) > 0, f"ERROR: Failed to predict for {dict_submodels[sub_model]}"
		list_tms.append(tms)
	
	# Combine embeddings
	list_tms = np.array(list_tms).T
	tms_features = np.vstack(list_tms)
	seq_features = np.vstack(df_embed['esm1b'].values)
	
	# Predict with ensemble
	fp_model_ensemb = os.path.join(folder_saved_models, name_ens_model + '.joblib')
	with open(fp_model_ensemb,"rb") as f:
		ensemb_model = joblib.load(f)
	pred = ensemb_model.predict(tms_features)

	return pred


def NanoMeltPredPipe(seq_records: list,
			 		do_align:bool, ncpus: int,  
					batch_size:int = 1024, verbose:bool = False) -> pd.DataFrame:
	
	'''Run the full melting temperature prediciton pipeline (embeddding and prediction).
	
	Parameters
    ----------
        - seq_records: list
            List of SeqRecords from the BioPython package. seq = str(record.seq) / id = record.id
		- do_align: bool
			If True, will align all the sequences using ANARCI (as implemented in AbNatiV).
		- ncpus: int
			Number of CPUs to use to paralelise the alignement and the embedding step of the dataset.
		- batch_size: int
			Batch size used to compute the embeddings on the sequences. 
			To avoid filling up all the memory with the embeddings
			when scoring a lot of sequences.
		- verbose: bool
			If True, will print more information about every step.

    Returns
    ----------
        - A dataframe with the sequences and the corresponding predicted temperatures
			and predicted errors. 
	
	'''
	# Attention order of keys matters in dict_submodels, the same order of predictions will be given to the ensemble model
	# NanoMelt models as in the original paper, Keys = name of saved submodel in folder_saved_models / Values = name of the embedding needed as inputs.

	# NanoMelt final trained model  as in the original paper
	with pkg_resources.path('nanomelt.model.saved_models', 'NanoMelt_finalmodel') as model_path:
		folder_saved_models = str(model_path)

	dict_submodels = {
		'SVR_esm1b': 'esm1b',
		'SVR_esm2_t30': 'esm2_t30',
		'GPR_vhse': 'vhse',
		'GPR_onehot': 'onehot',
					}

	# Batch over seq_records not to overload memory usage
	all_predictions = list()
	for batch_idx in tqdm(range(0, len(seq_records), batch_size), disable=not verbose):
		# Embed fasta
		batch_end = min(batch_idx + batch_size, len(seq_records))
		if verbose: print(f'-> Processing the sequences {batch_idx}-{batch_end}')

		batch_seq_records = seq_records[batch_idx:batch_end]
		df_embed = embed_fasta(batch_seq_records, set(dict_submodels.values()), batch_size, 
						 do_align=do_align, ncpus=ncpus, verbose=verbose)

		# Predict
		predictions = predict(df_embed, dict_submodels, 'ridge_ensemb', folder_saved_models)
		all_predictions.extend(predictions)

	all_predictions = [round(t, 2) for t in all_predictions]
	
	# Load in .csv
	data = {'ID': df_embed['id'], 'Aligned Sequence': df_embed['al_seq'], 
		 'Sequence': df_embed['seq'], "NanoMelt Tm (C)": all_predictions}

	return pd.DataFrame(data)




"""
 Copyright 2024. Aubin Ramon and Pietro Sormanni. CC BY-NC-SA 4.0
 
"""

import numpy as np
import pandas as pd

from sklearn.model_selection import GridSearchCV, BaseCrossValidator
from sklearn.preprocessing import StandardScaler
from sklearn.feature_selection import SelectFromModel, RFECV
from sklearn.pipeline import Pipeline
from sklearn.decomposition import PCA
from sklearn.linear_model import Ridge
from sklearn.model_selection import KFold

from typing import Tuple

from .regs import regression_dict
from .utils import perf_metrix


class CustomStratifiedKFold(BaseCrossValidator):
    """
    Stratified K-Folds cross-validator.

    Provides train/test indices to split data in train/test sets.

    This cross-validation object is a variation of KFold that returns
    stratified folds. The folds are made by preserving the percentage of
    samples for each class. It's a variation of StratifiedKFold that
    takes into account groups as group label and not y. 

    Parameters
    ----------
    - n_splits : int, default=5

    - shuffle : bool, default=False
        Whether to shuffle each class's samples before splitting into batches.
        Note that the samples within each split will not be shuffled.

    - random_state : int, default=0
        Pass an int for reproducible output across multiple function calls.
    """

    def __init__(self, n_splits=5, shuffle=True, random_state=0):
        self.n_splits = n_splits
        self.shuffle = shuffle
        self.random_state = random_state

    def get_n_splits(self, X=None, y=None, group=None):
        """
        Returns the number of splitting iterations in the cross-validator.
        
        Parameters
        --------
        - X: array-like, shape (n_samples, n_features), optional
            Input data, not used in this implementation.
        - y: array-like, shape (n_samples,), optional
            Target variable, not used in this implementation.
        - group: array-like, shape (n_samples,), optional
            Separate set of labels for stratification.

        Returns
        --------
        - n_splits: int
            The number of splitting iterations in the cross-validator.
        """
        return self.n_splits
    
    def _make_test_folds(self, X, y, groups=None):
        # Convert lists to NumPy arrays if they are not already
        X = np.array(X)
        y = np.array(y)
        groups = np.array(groups) if groups is not None else None

        # Ensure that X, y, and group have the same length
        if len(X) != len(y):
            raise ValueError("Input data and target variable must have the same length.")
        if groups is not None and len(groups) != len(X):
            raise ValueError("Group labels must have the same length as input data and target variable.")

        # Use group labels for stratification if provided, otherwise use target variable y
        labels = groups if groups is not None else y

        _, y_idx, y_inv = np.unique(labels, return_index=True, return_inverse=True)
        # y_inv encodes y according to lexicographic order. We invert y_idx to
        # map the classes so that they are encoded by order of appearance:
        # 0 represents the first label appearing in y, 1 the second, etc.
        _, class_perm = np.unique(y_idx, return_inverse=True)
        y_encoded = class_perm[y_inv]

        n_classes = len(y_idx)
        y_counts = np.bincount(y_encoded)
        min_groups = np.min(y_counts)
        if np.all(self.n_splits > y_counts):
            raise ValueError(
                "n_splits=%d cannot be greater than the"
                " number of members in each class." % (self.n_splits)
            )
        if self.n_splits > min_groups:
            print(
                "The least populated class in y has only %d"
                " members, which is less than n_splits=%d."
                % (min_groups, self.n_splits),
                UserWarning,
            )

        # Determine the optimal number of samples from each class in each fold,
        # using round robin over the sorted y. (This can be done direct from
        # counts, but that code is unreadable.)
        y_order = np.sort(y_encoded)
        allocation = np.asarray(
            [
                np.bincount(y_order[i :: self.n_splits], minlength=n_classes)
                for i in range(self.n_splits)
            ]
        )

        # Get class distributions
        uniques, counts = np.unique(labels, return_counts=True)
        props_ori = {unique: round(count/len(labels)*100,2) for unique, count in zip(uniques, counts)}
        #print('\n -> ORIGINAL DATA DISTRIBUTION (in %):', props_ori, '\n')

        # To maintain the data order dependencies as best as possible within
        # the stratification constraint, we assign samples from each class in
        # blocks (and then mess that up when shuffle=True).
        test_folds = np.empty(len(labels), dtype="i")
        for k in range(n_classes):
            # since the kth column of allocation stores the number of samples
            # of class k in each test set, this generates blocks of fold
            # indices corresponding to the allocation for class k.
            folds_for_class = np.arange(self.n_splits).repeat(allocation[:, k])
            if self.shuffle:
                np.random.seed(self.random_state)
                np.random.shuffle(folds_for_class)

            test_folds[y_encoded == k] = folds_for_class

        return test_folds

    def _iter_test_masks(self, X, y=None, groups=None):
        test_folds = self._make_test_folds(X, y, groups)
        for i in range(self.n_splits):
            test_fold_i = test_folds == i

            # Get class distributions
            if groups is not None:
                test_labels = groups[test_fold_i]
                uniques, counts = np.unique(test_labels, return_counts=True)
                new_props = {unique: round(count/len(test_labels)*100,2) for unique, count in zip(uniques, counts)}
                #print(f'\n -> TEST FOLD-{i} DATA DISTRIBUTION (in %):', new_props, '\n')

            yield test_fold_i


    def split(self, X, y, groups=None):
        """Generate indices to split data into training and test set.

        Parameters
        ----------
        X : array-like of shape (n_samples, n_features)
            Training data, where `n_samples` is the number of samples
            and `n_features` is the number of features.

            Note that providing ``y`` is sufficient to generate the splits and
            hence ``np.zeros(n_samples)`` may be used as a placeholder for
            ``X`` instead of actual training data.

        y : array-like of shape (n_samples,)
            The target variable for supervised learning problems.
            Stratification is done based on the y labels.

        Yields
        ------
        train : ndarray
            The training set indices for that split.

        test : ndarray
            The testing set indices for that split.

        Notes
        -----
        Randomized CV splitters may return different results for each call of
        split. You can make the results identical by setting `random_state`
        to an integer.
        """

        return super().split(X, y, groups)


class CustomLeaveOneClassOut(BaseCrossValidator):
    """
    Leave-One-Out cross-validator for each class.

    Provides train/test indices to split data in train/test sets, ensuring each class
    is left out in turn.

    """

    def __init__(self):
        return

    def get_n_splits(self, X=None, y=None, groups=None):
        """
        Returns the number of splitting iterations in the cross-validator.
        
        Parameters
        ----------
        - X: array-like, shape (n_samples, n_features), optional
            Input data, not used in this implementation.
        - y: array-like, shape (n_samples,), optional
            Target variable for stratification.
        - groups: array-like, shape (n_samples,), optional
            Group labels for stratification.
        
        Returns
        -------
        - n_splits: int
            The number of splitting iterations in the cross-validator.
        """
        return len(np.unique(groups))

    def split(self, X, y, groups):
        """
        Generate indices to split data into training and test set.
        
        Parameters
        ----------
        - X: array-like, shape (n_samples, n_features)
            The input data.
        - y: array-like, shape (n_samples,)
            The target variable.
        - groups: array-like, shape (n_samples,)
            Group labels for stratification.
        
        Returns
        -------
        - train: ndarray
            The training set indices for that split.
        - test: ndarray
            The testing set indices for that split.
        """
        unique_groups = np.unique(groups)
        for group in unique_groups:
            test_indices = np.where(groups == group)[0]
            train_indices = np.where(groups != group)[0]
            yield train_indices, test_indices


class RegressionFit():

    '''
    For given train and test data and target, fit to the given regression.
    Hyperparameters searched by GridSearchCV with StratifiedKFold using customised labels (e.g. Tm methods).

    Attributes
    --------
    - data_train, target_train, label_train: data and target should be np.ndarray;

    - data_test, target_test, label_test: data and target should be np.ndarray;

    - feature_sel_method: str, default = None
        Any from ['SelectFromModel', 'RFECV', 'PCA', None];

    - GridSearch_folds: int, default = 5
        num of folds performed by GridSearchCV with the train_set;

    - random_state: int, default = 5
        Randome state used for StratifiedKFold. 
        Pass an int for reproducible output across multiple function calls;
    
    - regression: str, default = 'ridge'
        Regression model to fit to the dataset.
    '''

    def __init__(self, data_train, target_train, label_train,
                feature_sel_method = None,
                GridSearch_folds: int = 5,
                random_state: int = 5,
                regression: str = 'ridge',
                strat_method : str = 'class_tm_method'):
        
        self.data_train = data_train
        self.target_train = target_train
        self.label_train = label_train

        self.feature_sel_method = feature_sel_method
        self.GridSearch_folds = GridSearch_folds
        self.random_state = random_state
        self.regression = regression
        self.strat_method = strat_method

        if self.feature_sel_method == None:
            self.pipe = Pipeline([('scale', StandardScaler()),],verbose=False)
            
        else:
            if self.feature_sel_method == 'PCA':
                feature_sel = PCA(n_components=0.6)

            # Feature selection with ridge as default
            if self.feature_sel_method == 'SelectFromModel':
                feature_sel = SelectFromModel(Ridge(alpha=200), prefit=False)
            
            if self.feature_sel_method == 'RFECV':
                feature_sel = RFECV(Ridge(alpha=200), step=0.1, cv=3)

            self.pipe = Pipeline([('scale', StandardScaler()),
                            ('reduce_dims', feature_sel),],verbose=False)

    def param_scores(self, ):
        """
        Fit the given regression to the train dataset.
        Hyperparameters searched by GridSearchCV with StratifiedKFold using customised labels (e.g. Tm methods).

        Returns
        --------
        - param_scores: parameter performance dict from grid_search.cv_results_ 
            (see sklearn.model_selection.GridSearchCV documentation)

        - pipe: sklearn.pipeline.Pipeline object

        Example
        --------
        Example pipe: Pipeline(steps=[
            ('scale', StandardScaler()),
            ('reduce_dims', SelectFromModel(estimator=Ridge(alpha=200))), # preprocessing steps
            ['regr', Ridge(alpha=200)] # regression model with best params from GridSearchCV
            ])
        """
        # Define regression model and param_grid
        regr_model = regression_dict[f'{self.regression}'][0]
        param_grid = regression_dict[f'{self.regression}'][1]
        prefix = 'regr__' # update param_grid format for pipe
        param_grid_updated = {prefix + key: value for key, value in param_grid.items()}
        self.pipe.steps.append(['regr',regr_model])

        # HP tuning with GridSearchCV on the train set
        # default scoring method in GridSearch = default scoring method for the regression; 
        # define scoring to be neg MSE for consistency; negative such that max score gives best model
        if self.strat_method == 'none':
            grid_search = GridSearchCV(self.pipe, param_grid=param_grid_updated, cv=self.GridSearch_folds, scoring='neg_mean_squared_error', n_jobs=-1)
        
        elif self.strat_method == 'loco':
            skf = CustomLeaveOneClassOut()
            grid_search = GridSearchCV(self.pipe, param_grid=param_grid_updated, cv=skf, scoring='neg_mean_squared_error', n_jobs=-1)

        else:
            skf = CustomStratifiedKFold(n_splits=self.GridSearch_folds, shuffle=True, random_state=self.random_state)
            grid_search = GridSearchCV(self.pipe, param_grid=param_grid_updated, cv=skf, scoring='neg_mean_squared_error', n_jobs=-1)

        grid_search.fit(self.data_train, self.target_train, groups = self.label_train)

        # Update best model params
        param_scores = grid_search.cv_results_
        pipe = self.pipe.set_params(**grid_search.best_params_)
        
        return param_scores, pipe
    


def StratifiedNestedCV(data: list, target: list, label: list,
                    num_outer_fold: int = 3,
                    num_inner_fold: int = 3,
                    random_states: list = [0,42,100],
                    regression: str = 'ridge',
                    feature_sel_method: str = 'SelectFromModel',
                    strat_method: str = 'expmethod_kmeans',
                    return_averaged: bool = True,
                    ) -> Tuple[list,list,list]:
    """
    Perform stratified nested CV on the dataset provided.
    Labelling of the groups for stratification is provided through the list "label".  
    "strat_method" tells the program which type of stratification algo to do and save the results as.
        e.g., 'none', 'loco' (leave one class out), or other for full stratification(like 'expmethod_kmeans') 

    If multiple seeds are given, the models will generate the same amount of predictions for each input at each seed repeat
    in list_all_seed_predictions

    Parameters
    --------
    - data: list
        List of inputs
    - target: list
        List of target for the regressions
    - label: list
        List of labels for stratificaiton 
    - num_outer_fold: int
    - num_inner_fold: int
    - random_states: list of int
        Random seeds for multiple repeats, only put one if you do not want repeats
    - regression: str
        regression name to be used (e.g., 'linear','lasso','elasticnet'). See regression_dict in regs.py
    - feature_sel_method: str, default = None
        Any from ['SelectFromModel', 'RFECV', 'PCA', None]
    - strat_method: str
        "strat_method" tells the program which type of stratification algo to do and save the results as.
        e.g., 'none', 'loco' (leave one class out), or other for full stratification(like 'expmethod_kmeans') 
    return_averaged: bool
        If True, will average all the performances over the seeds.

    Returns
    --------
    - train_metrics: if return_averaged, performances will have been averaged each outer layer and each seed, 
        otherwise will save every single perf (will need to mean later then, usefull in stacking.py)
    - test_metrics: same but for test
    - list_all_seed_predictions: list of predictions for each seed

    """
    print(f"==================== Stratified Nested CV with {regression} ====================")
    
    list_all_seed_pearson_train,list_all_seed_spearman_train,list_all_seed_mae_train, list_all_seed_std_ratio_train = [],[],[],[]
    list_all_seed_pearson_p_train,list_all_seed_spearman_p_train = [],[]
    
    list_all_seed_pearson_test,list_all_seed_spearman_test,list_all_seed_mae_test, list_all_seed_std_ratio_test = [],[],[],[]
    list_all_seed_pearson_p_test,list_all_seed_spearman_p_test = [],[]
    
    list_all_seed_predictions = []

    for outer_seed in random_states:
        print(f"\n------------ Seed number = {outer_seed} ------------")
        
        # Outer loop
        list_outer_train_pearson,list_outer_train_spearman,list_outer_train_mae, list_outer_train_std_ratio = [],[],[],[]
        list_outer_train_pearson_p,list_outer_train_spearman_p = [],[]

        list_outer_test_pearson,list_outer_test_spearman,list_outer_test_mae, list_outer_test_std_ratio = [],[],[],[]
        list_outer_test_pearson_p,list_outer_test_spearman_p = [],[]

        list_outer_test_pred = [None]*len(data)

        if strat_method == 'none':
            outer_splits = KFold(n_splits=num_outer_fold, shuffle=True, random_state=outer_seed)
        elif strat_method == 'loco':
            outer_splits = CustomLeaveOneClassOut()
        else:
            outer_splits = CustomStratifiedKFold(n_splits=num_outer_fold, shuffle=True, random_state=outer_seed)

        for fold, (train_index, test_index) in enumerate(outer_splits.split(data, target, groups=label)):
            print(f'\n ===== Fold {fold+1} =====\n')

            data_train, data_test = np.vstack([data[i] for i in train_index]), np.vstack([data[i] for i in test_index])
            target_train, target_test = np.array([target[i] for i in train_index]), np.array([target[i] for i in test_index])
            label_train, label_test = np.array([label[i] for i  in train_index]), np.array([label[i] for i in test_index])

            # Inner loop
            list_score_dict = list()
            for inner_seed in random_states:
                inner_loop = RegressionFit(data_train, target_train, label_train,
                            feature_sel_method = feature_sel_method, GridSearch_folds = num_inner_fold, 
                            random_state = inner_seed, regression = regression, strat_method = strat_method)
                param_scores, pipe = inner_loop.param_scores()
                list_score_dict.append(param_scores)
            
            # Store params scores and mean over all seeds
            list_mean_test_score= list()
            for score_dict in list_score_dict:
                list_mean_test_score.append(np.array(score_dict['mean_test_score']))
            inner_mean_test_score = [np.mean(k) for k in zip(*list_mean_test_score)]
            
            # Rank params according to mean_test_score with the highest score ranked as 1
            sorted_indices = np.argsort(inner_mean_test_score)[::-1]
            rank_list = np.empty_like(sorted_indices)
            rank_list[sorted_indices] = np.arange(1, len(inner_mean_test_score) + 1)
            
			# Collect all trialed params in GridSearch into param_dict
            # Same params tried across all seeds, so only one score_dict is used to access the params
            # Also save the corresponding mean_test_score across all seeds and the ranking
            param_dict = {}
            param_dict['params'] = list_score_dict[0]['params']
            param_dict['mean_test_score'] = inner_mean_test_score
            param_dict['rank_test_score'] = rank_list
            
			# Find best params across all seeds
            param_df = pd.DataFrame(param_dict)
            best_params = param_df.loc[param_df['rank_test_score']==1, 'params'].values[0]
            pipe.set_params(**best_params)
            print(f"\nTuned params for outer fold {fold+1} with seed {outer_seed}: \n{pipe}")
            
			# Train with the best model
            pipe.fit(data_train, target_train)
            
			# Save train and test performance 
            pred_train = pipe.predict(data_train) # Outer Train
            pred_test = pipe.predict(data_test) # Outer Test

            train_pearson, train_pearson_p, train_spearman, train_spearman_p, train_mae, train_std_ratio = perf_metrix(pred_train,target_train)
            test_pearson, test_pearson_p, test_spearman, test_spearman_p, test_mae, test_std_ratio = perf_metrix(pred_test,target_test)

            list_outer_train_pearson.append(train_pearson)
            list_outer_train_pearson_p.append(train_pearson_p)
            list_outer_train_spearman.append(train_spearman)
            list_outer_train_spearman_p.append(train_spearman_p)
            list_outer_train_mae.append(train_mae)
            list_outer_train_std_ratio.append(train_std_ratio)

            list_outer_test_pearson.append(test_pearson)
            list_outer_test_pearson_p.append(test_pearson_p)
            list_outer_test_spearman.append(test_spearman)
            list_outer_test_spearman_p.append(test_spearman_p)
            list_outer_test_mae.append(test_mae)
            list_outer_test_std_ratio.append(test_std_ratio)

			# Save predictions
            for i in range(len(test_index)):
                test_id = test_index[i]
                prediction = pred_test[i]
                list_outer_test_pred[test_id] = prediction

        # Mean over all outer folds and save it into the seed list
        seed_fin_pearson_train = list_outer_train_pearson
        seed_fin_pearson_p_train = list_outer_train_pearson_p
        seed_fin_spearman_train = list_outer_train_spearman
        seed_fin_spearman_p_train = list_outer_train_spearman_p
        seed_fin_mae_train = list_outer_train_mae
        seed_fin_std_ratio_train = list_outer_train_std_ratio

        seed_fin_pearson_test = list_outer_test_pearson
        seed_fin_pearson_p_test = list_outer_test_pearson_p
        seed_fin_spearman_test = list_outer_test_spearman
        seed_fin_spearman_p_test = list_outer_test_spearman_p
        seed_fin_mae_test = list_outer_test_mae
        seed_fin_std_ratio_test = list_outer_test_std_ratio
        print(f"\nTest spearman (outer loop) = {np.mean(seed_fin_spearman_test)} for seed {outer_seed}")

        list_all_seed_pearson_train.append(seed_fin_pearson_train)
        list_all_seed_pearson_p_train.append(seed_fin_pearson_p_train)
        list_all_seed_spearman_train.append(seed_fin_spearman_train)
        list_all_seed_spearman_p_train.append(seed_fin_spearman_p_train)
        list_all_seed_mae_train.append(seed_fin_mae_train)
        list_all_seed_std_ratio_train.append(seed_fin_std_ratio_train)

        list_all_seed_pearson_test.append(seed_fin_pearson_test)
        list_all_seed_pearson_p_test.append(seed_fin_pearson_p_test)
        list_all_seed_spearman_test.append(seed_fin_spearman_test)
        list_all_seed_spearman_p_test.append(seed_fin_spearman_p_test)
        list_all_seed_mae_test.append(seed_fin_mae_test)
        list_all_seed_std_ratio_test.append(seed_fin_std_ratio_test)

        list_all_seed_predictions.append(list_outer_test_pred)

    # Mean over all outer seeds
    if return_averaged:
        fin_pearson_train = np.mean(list_all_seed_pearson_train)
        fin_pearson_train_p = np.mean(list_all_seed_pearson_p_train)
        std_pearson_train = np.std(list_all_seed_pearson_train)
        fin_spearman_train = np.mean(list_all_seed_spearman_train)
        fin_spearman_train_p = np.mean(list_all_seed_spearman_p_train)
        std_spearman_train = np.std(list_all_seed_spearman_train)
        fin_mae_train = np.mean(list_all_seed_mae_train)
        std_mae_train = np.std(list_all_seed_mae_train)
        fin_std_ratio_train = np.mean(list_all_seed_std_ratio_train)
        std_std_ratio_train = np.std(list_all_seed_std_ratio_train)

        fin_pearson_test = np.mean(list_all_seed_pearson_test)
        fin_pearson_test_p = np.mean(list_all_seed_pearson_p_test)
        std_pearson_test = np.std(list_all_seed_pearson_test)
        fin_spearman_test = np.mean(list_all_seed_spearman_test)
        fin_spearman_test_p = np.mean(list_all_seed_spearman_p_test)
        std_spearman_test = np.std(list_all_seed_spearman_test)
        fin_mae_test = np.mean(list_all_seed_mae_test)
        std_mae_test = np.std(list_all_seed_mae_test)
        fin_std_ratio_test = np.mean(list_all_seed_std_ratio_test)
        std_std_ratio_test = np.std(list_all_seed_std_ratio_test)
    
        train_metrics = [fin_pearson_train, fin_pearson_train_p, std_pearson_train, 
                        fin_spearman_train, fin_spearman_train_p, std_spearman_train, 
                        fin_mae_train, std_mae_train, fin_std_ratio_train, std_std_ratio_train]
        test_metrics = [fin_pearson_test, fin_pearson_test_p, std_pearson_test, 
                        fin_spearman_test, fin_spearman_test_p, std_spearman_test, 
                        fin_mae_test, std_mae_test, fin_std_ratio_test, std_std_ratio_test]
    
        return [round(x,3) for x in train_metrics], [round(x,3) for x in test_metrics], list_all_seed_predictions
    
    else: 
        train_metrics = [list_all_seed_pearson_train, [], [], list_all_seed_spearman_train, [], [],
                        list_all_seed_mae_train, [], list_all_seed_std_ratio_train, []]
        test_metrics = [list_all_seed_pearson_test, [], [], list_all_seed_spearman_test, [], [],
                        list_all_seed_mae_test, [], list_all_seed_std_ratio_test, []]
    
        return train_metrics, test_metrics, list_all_seed_predictions
    
     

